#!/usr/bin/perl
#
# $Revision: 1.19 $
# Copyright 2008 Teleflora
#
# Utility for killing off user terminals.
#
use strict;
use warnings;
use Getopt::Long;
use English;


my $HELP = 0;
my $VERSION = 0;
my $NOW = 0;
my $ALL = 0;
my $BBTERM = 1;
my $LIST = 0;
my @PIDS = ();
my %SESSION = ();
my $returnval = 0;




GetOptions(
	"version" =>  \$VERSION,
	"help" => \$HELP,
	"now" => \$NOW,
	"all" => \$ALL,
	"list" => \$LIST,
	"bbterm!" => \$BBTERM,
);


# --help
if($HELP != 0) {
	usage();
	exit(0);
}

# --version
if($VERSION != 0) {
	print("$0 " . '$Revision: 1.19 $' . "\n");
	exit(0);
}


# Are we root?
if($UID != 0 || $EUID != 0) {
	print("Error: Must be sudo to run this utility.\n");
	exit(1);
}



# Who is logged in? (based on commandline parameters we specified)
populate_sessions(\@ARGV);

# --list
if($LIST != 0) {
	exit(list_terms());
}


if($ALL == 0) {
	# If user didn't ask to kill "all", then, give them a UI to "pick"
	# which user to kill.
	@PIDS = pick_terms();
} else {
	# --all
	# means to kill all specified users.
	my $pid = "";
	foreach $pid (keys %SESSION) {
		push(@PIDS, $pid);
	}
}


if($#PIDS >= 0) {
	$returnval = confirm_kill(\@PIDS);
	if($returnval == 0) {
		killterm(\@PIDS);
	}
}

exit(0);
################################################################################
################################################################################
################################################################################


sub usage
{
	print("Usage:\n");
	print("$0 --help\n");
	print("$0 --version\n");
	print("\n");
	print("Options:\n");
	print("$0 --nobbterm # Show all terminals, not just BBx terms.\n");
	print("$0 --all # Kill all instances of a particular username.\n");
	print("$0 --now # Don't issue the user a warning.\n");

	print("\n");
	print("Example Usages:\n");
	print("sudo $0\n");
	print("sudo $0 user1 user2 ...\n");
	print("\n");
	print("Kill a single PID (or all pids) with no warning.\n");
	print("sudo $0 --now\n");
	print("sudo $0 --now user1 user2 ...\n");
	print("\n");
	print("Kill all sessions (of a specified user) with warning.\n");
	print("sudo $0 --all\n");
	print("sudo $0 --all user1 user2 ...\n");
	print("\n");
	print("Kill all specified users without any warning.\n");
	print("sudo $0 --now --all\n");
	print("sudo $0 --now --all user1 user2 ...\n");
}



#
# Given the PID of a controlling TTY, get the "BBTERM" environment
# variable of the underlying BBx/BBj process.
#
sub get_bbterm
{
	my $pid = $_[0];
	my @array = ();
	my $pidfound = -1;
	my $buffer = "";
	my $bbterm = "";

	
	# Use "pstree -p" to determine the PID of Pro5.
	# It is *that* PID which has the BBTERM environment
	# variable set.
	open(PSTREE, "pstree -p $pid |");
	while(<PSTREE>) {
		if(/(pro5[bs]\()(\d+)(\))/) {
			$pidfound = $2;
			last;
		}
		if(/(bbj[bs]\()(\d+)(\))/) {
			$pidfound = $2;
			last;
		}


	}
	close(PSTREE);

	if($pidfound > 0) {
		open(FILE, "sudo cat /proc/$pidfound/environ |");
		sysread(FILE, $buffer, 2048);
		close(FILE);
		@array=split(/\0/, $buffer);
		foreach(@array) {
			if(/(BBTERM=)(\S+)/) {
				$bbterm = $2;
				last;
			}
		}
	}

	return($bbterm);
}




sub populate_sessions
{
	my $pid = "";
	my $username = "";
	my $mytty = "";
	my $terminal = "";
	my $bbterm = "";
	my @array = ();
	my @userfilters = @{$_[0]};



	# This should always be an "option" for which users to kill.
	# "All" in the context of this script means "all users except self."
	if($#userfilters < 0) {
		$SESSION{'all'}{'pid'} = "all";
		$SESSION{'all'}{'user'} = "All Users";
		$SESSION{'all'}{'hostname'} = "Any Host";
		$SESSION{'all'}{'tty'} = "All Sessions";
		$SESSION{'all'}{'bbterm'} = "All Terminals";
	}

	# What TTY are we on? 
	# We want to exclude this from our kill list.
	open(PIPE, "tty |");
	while(<PIPE>) {
		chomp;
		$mytty .= $_;
	}
	close(PIPE);
	$mytty =~ s/\/dev\///g;


	# Get a list of who all sessions, possibly filltered.
	open(PIPE, "ps wwaux |");
	while(<PIPE>) {
		chomp;
		next until(/\/bin\/bash \-\-login/);

		@array = split(/\s+/);
		$username = $array[0];
		$pid = $array[1];
		$terminal = $array[6];

		# Don't allow us to kill ourselves.
		next if($terminal eq $mytty);

		# Get BBTERM from environment.
		$bbterm = get_bbterm($pid);
		next if($BBTERM != 0) && ($bbterm eq "");




		# If we specified usernames, then, 
		# only look for sessions by that user.
		if(@userfilters) {
			next unless (grep(/$username/, @userfilters));
		}

		# Information from "who"
		$SESSION{$pid}{'pid'} = $pid;
		$SESSION{$pid}{'user'} = $username;
		$SESSION{$pid}{'tty'} = $terminal;
		$SESSION{$pid}{'hostname'} = "";
		$SESSION{$pid}{'bbterm'} = $bbterm;
	}
	close(PIPE);
}





#
# Present a menu of terminals to be killed
#
sub pick_terms
{
	my %hash = ();
	my @array = ();
	my @chosenpids = ();
	my $pid = "";
	my $command = "";


	$command = "dialog";
	$command .= " --clear";
	$command .= " --title \"Kill RTI Terminal Sessions\"";
	$command .= " --menu \"Select the terminal session you wish to kill.\"";

	# Rows, Columns Visible Menu items
	$command .= " 20 65 15";

	# List all users.
	foreach $pid(sort keys(%SESSION)) {
		$pid =~ s/\s+//g;
		$command .= sprintf(" \"%-8.8s\" \"%-8.8s from %-8.8s (%s)\"",
				$SESSION{$pid}{'pid'},
				$SESSION{$pid}{'user'},
				$SESSION{$pid}{'tty'},
				$SESSION{$pid}{'bbterm'}
			);
	}





	# If we specified "--all" on the commandline, then, we don't need
	# a user interface to "pick" who to kill, we want to kill everyone
	# specified.
	# Thus
	# --all user1 user2 will kill all instances for user1 and user2
	# --all will kill all users expect the currently logged in user.
	if($ALL != 0) {
		return("all");
	}



	# Run user interface to pick which user to kill.
	print("$command\n");
	system("$command 2> /tmp/killterms.$$");
	open(FILE, "< /tmp/killterms.$$");
	while(<FILE>) {
		chomp;
		push(@chosenpids, int($_));
		last;
	}
	close(FILE);
	unlink("/tmp/killterms.$$");

	# User pressed cancel.
	if($#chosenpids < 0) {
		return(@chosenpids);
	}

	# If user picked "all" pids, then, return an array
	# of all pids in the "Session".
	if($chosenpids[0] eq "all") {
		foreach (keys %SESSION) {
			push(@chosenpids, $_);
		}
	}



	return(@chosenpids);
}



#
# Present a menu of terminals to be killed
#
sub list_terms
{
	my %hash = ();
	my @array = ();
	my @chosenpids = ();
	my $pid = "";
	my $command = "";


	# List all users.
	foreach $pid(sort keys(%SESSION)) {
		$pid =~ s/\s+//g;
		next if($pid =~ /all/);

		printf("\"%s\", \"%s\", \"%s\", \"%s\"\n",
				$SESSION{$pid}{'pid'},
				$SESSION{$pid}{'user'},
				$SESSION{$pid}{'tty'},
				$SESSION{$pid}{'bbterm'}
			);
	}

	return(0);
}





#
# Give more info about what is to be killed.
#
sub confirm_kill
{
	my @pidschosen = @{$_[0]};
	my $pid = "";
	my $proctree = "";
	my $command = "";


	# We are killing a single pid.
	$command = "dialog";
	$command .= " --clear";
	if($#pidschosen == 0) {
		$pid = $pidschosen[0];
		$command .= " --title \"Kill Process $pid?\"";
		$command .= " --yesno \"";
		open(PIPE, "pstree -a $pid |");
		while(<PIPE>) {
			chomp;
			$proctree .= "$_\\n";
		}
		close(PIPE);

		# Escape out backtick chars.
		$proctree =~ s/\`/\\`/g;

		$command .= "Current User: $SESSION{$pid}{'user'}\\n";
		$command .= "From Host: $SESSION{$pid}{'hostname'}\\n";
		$command .= "Using TTY: $SESSION{$pid}{'tty'}\\n";
		$command .= "BBx Terminal: $SESSION{$pid}{'bbterm'}\\n";
		$command .= "Running Programs:\\n";
		$command .= $proctree;


	# A list of pids was specified.
	} elsif ($#pidschosen > 0) {
		$command .= " --title \"Kill Multiple Processes ?\"";
		$command .= " --yesno \"";

		$command .= "\\n";
		foreach $pid (@pidschosen) {


			$command .= sprintf("%-8.8s %-8.8s from %-8.8s (%s)\\n",
				$SESSION{$pid}{'pid'},
				$SESSION{$pid}{'user'},
				$SESSION{$pid}{'tty'},
				$SESSION{$pid}{'bbterm'}
				);
		}
	}
	$command .= "\\n";
	$command .= "\\n";
	$command .= "\"";
	$command .= " 20 60";



	system("$command");
	return($?);
}


#
# Actually kill a termainal hre.
#
sub killterm
{
	my @pidschosen = @{$_[0]};
	my @array = ();
	my $pid = "";
	my $i = 0;
	my $command = "";
	my $elapsed_time = 0;
	my $warnwait = 0;


	# Send the user a "You are about to be killed" message.
	if($NOW == 0) {

		$elapsed_time = 0;
		$warnwait = 15;
		foreach $warnwait(15, 10, 5, 1) {

			foreach $pid (@pidschosen) {
				# Send the victim a warning.
				$command = "| sudo write ";
				$command .= " $SESSION{$pid}{'user'}";
				$command .= " $SESSION{$pid}{'tty'}";

				open (PIPE,"$command");
					print PIPE "\n";
					print PIPE "*** Your session will be killed in " . (30 - $elapsed_time) . " seconds *** \n";
				close(PIPE);
			}


			# Tell user what is going on, 
			# and then sleep
			$command = "dialog";
			$command .= " --clear";
			$command .= " --title \"Sending User Notification\"";
			$command .= " --sleep $warnwait";
			$command .= " --infobox";
			$command .= " \"Warning users about impending session closure in";
			$command .= " " . (30 - $elapsed_time) . " seconds.\\n";
			foreach $pid (@pidschosen) {
				$command .= " $SESSION{$pid}{'user'} Killing Terminal in " . (30 - $elapsed_time) . " seconds...";
			}
			$command .= "\" 10 35";
			system($command);


			$elapsed_time += $warnwait;
		}
	}

	foreach $pid(@pidschosen) {
		system("sudo kill -HUP $pid");
	}
	sleep(1);
	foreach $pid(@pidschosen) {
		system("sudo kill -TERM $pid 2> /dev/null");
	}
	sleep(1);
	foreach $pid(@pidschosen) {
		system("sudo kill -KILL $pid 2> /dev/null");
	}

	return(0);
}
